from typing import List

import typer
from typing_extensions import Annotated

from pip2spack.actions.create.create_action import create_action
from pip2spack.actions.dowload.download_action import download_action
from pip2spack.actions.update.update_action import update_action
from pip2spack.framework.helpers import get_spack_repository

app = typer.Typer()

PACKAGE_TYPE = Annotated[
    List[str], typer.Argument(help="Name of the package at pypi.org")
]


@app.command(help="Create a new package")
def create(names: PACKAGE_TYPE) -> None:
    get_spack_repository()
    create_action(names=names)


@app.command(help="Update an existing package name")
def update(names: PACKAGE_TYPE):
    get_spack_repository()
    update_action(names=names)


@app.command(help="Download the latest package from pypi.org")
def download(names: PACKAGE_TYPE):
    get_spack_repository()
    download_action(names=names)


@app.command(help="Show version")
def version() -> None:
    print("pip2spack: v2.0.0a1")


if __name__ == "__main__":
    app()



from pip2spack.core.pypi_package import PyPiPackage


class TestPyPiPackage:
    def test_pypi_package__generate_correct_path_for_package_with_dash_in_name(
        self, mock_pypi_configshell_fb_response: None
    ) -> None:
        """Base on the issue #16 https://github.com/NexSabre/pip2spack/issues/16"""
        assert (
            PyPiPackage("configshell-fb").url
            == "https://pypi.io/packages/source/c/configshell-fb/configshell-fb-1.1.29.tar.gz"
        )

    def test_pypi_package__verify_package_name_builder__configshell_fb(
        self, mock_pypi_configshell_fb_response: None
    ):
        assert PyPiPackage("configshell-fb").package_name_builder() == "PyConfigshellFb"

    def test_pypi_package__verify_package_name_builder__jsl(
        self, mock_pypi_jsl_response: None
    ) -> None:
        assert PyPiPackage("jsl").package_name_builder() == "PyJsl"

    def test_pypi_package__package_name_builder_should_replace_dash_in_the_info_name(
        self, mock_pypi_jsl_response: None
    ) -> None:
        package = PyPiPackage("jsl")
        package.content["info"]["name"] = "jsl-up"

        assert package.package_name_builder() == "PyJslUp"

    def test_pypi_package__test_generate_data_return_message(
        self, mock_pypi_jsl_response: None
    ) -> None:
        return_message = PyPiPackage("jsl")._generate_data()
        for k, v in return_message.items():
            if k == "depends_on_setuptools":
                continue
            assert v, "Should be True"

    def test_pypi_package__ctypesgen_should_be_categorized_as_not_source(
        self, mock_pypi_ctypesgen_response: None
    ) -> None:
        assert not PyPiPackage(
            "ctypesgen"
        ).source, (
            "Package should be tread as wheel package, source should be unavailable"
        )



import json
from typing import Any, Dict, List

import requests
from jinja2 import Environment, FileSystemLoader, PackageLoader

from pip2spack.framework.messages import Messages


class PyPiPackage:
    content: Dict = {}
    versions: List = []
    url: str = ""
    homepage: str = ""
    maintainers: List = []
    package_name: str = ""
    summary: str = ""
    source: bool = False

    github: str = ""
    pypi: str = ""

    def __init__(self, content):
        self.content = self._download_missing_information(content)

        self._url(source=self.__detect_approach())

        if self.source:
            self._versions_formatter()
        else:
            self._wheel_versions_formatter()
        self.version_builder()

        self.package_name_builder()
        self.summary_builder()
        self.homepage_builder()

        self._fill_pypi()

    @property
    def _get_versions(self):
        self.versions = []
        self._versions_formatter()
        self.version_builder()
        return self.versions

    def __detect_approach(self) -> bool:
        return True if self.content["urls"][0]["url"].endswith(".tar.gz") else False

    def _fill_pypi(self):
        sorted_versions: List[str] = sorted([str(x) for x in self.content["releases"].keys()], key=lambda x: x.split("."), reverse=False)
        for version in sorted_versions:
            for x in self.content["releases"][version]:
                if x["packagetype"] == "sdist":
                    self.pypi = f"{self.content['info']['name']}/{x['filename']}"
                    break


    @staticmethod
    def _download_missing_information(package_name: str) -> Dict[str, Any]:
        if package_name.startswith("py-"):
            package_name = package_name.replace("py-", "")
        status = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        if status.status_code != 200:
            print("Package doesnt not exist at PyPi.org")
            raise
        return json.loads(status.content)

    def _url(self, source):
        def _generate_pypi_uri(filename_with_version):
            if not filename_with_version:
                raise Exception("Can not generate a proper master link to the package")
            return (
                f"https://pypi.io/packages/source/{filename_with_version[0]}/"
                f'{"-".join(filename_with_version.split("-")[:-1])}/{filename_with_version}'
            )

        if source:
            for url in self.content["urls"]:
                if url["url"].endswith(".tar.gz"):
                    self.url = _generate_pypi_uri(url["url"].split("/")[-1])
                    self.source = True
                    break
        # else:
        #     # preferable packages are .tar.gz, but some author does not include it, they just provide .whl package
        #     for u in self.content["urls"]:
        #         if u["url"].endswith(".whl"):
        #             self.url = u["url"]
        #             self.source = False
        #             break

        # if not self.url:
        #     raise Exception("Source package was not found")
        #
        # elif self.url and not self.source:
        #     print("[Experimental] Using a wheel package...")



    def _homepage(self):
        self.homepage = self.content["home_page"]

    def _versions_formatter(self):
        for version, version_content in self.content["releases"].items():
            only_sdist_version = [
                x
                for x in version_content
                if (x["url"].endswith(".tar.gz") and x["packagetype"] == "sdist")
            ]
            if not only_sdist_version:
                continue

            for proper_version in only_sdist_version:
                try:
                    # TODO Not support .whl packages yet. Only source code in .tar.gz
                    if not proper_version["filename"].endswith(".tar.gz"):
                        continue
                    self.versions.append(
                        (str(version), str(proper_version["digests"]["sha256"]))
                    )
                except IndexError:
                    continue

    def _wheel_versions_formatter(self):
        # TODO: Experimental support of the .whl. It can be pulled out if creates more cons than props
        for version, version_content in self.content["releases"].items():
            only_bdist_wheel_version = [
                x
                for x in version_content
                if (x["url"].endswith(".whl") and x["packagetype"] == "bdist_wheel")
            ]

            for proper_version in only_bdist_wheel_version:
                try:
                    if not proper_version["filename"].endswith(".whl"):
                        continue
                    self.versions.append(
                        (str(version), str(proper_version["digests"]["sha256"]))
                    )
                except IndexError:
                    continue

    def version_builder(self) -> str:
        raw_versions: str = ""
        for v in self.versions:
            raw_versions += f'version("{str(v[0])}", sha256="{str(v[1])}")\n'
        return raw_versions

    def package_name_builder(self) -> str:
        name = self.content["info"]["name"]
        if name[0].islower():
            name = name[0].upper() + name[1:]

        l_name = list(name)
        for i in [index for index, element in enumerate(l_name) if element == "-"]:
            try:
                if l_name[i + 1].islower():
                    l_name[i + 1] = l_name[i + 1].upper()
            except Exception as e:
                Messages.error(f"package_name_builder: {e}")
            finally:
                name = "".join(l_name).replace("-", "")

        if not name.startswith("Py"):
            name = "Py" + name
        self.package_name = name
        return name

    def summary_builder(self):
        self.summary = self.content["info"]["summary"]

    def homepage_builder(self):
        self.homepage = self.content["info"]["project_urls"]["Homepage"]

    def _generate_data(self) -> dict:
        return {
            "package_name": self.package_name,
            "summary": self.summary,
            "homepage": self.homepage,
            "url": self.url,
            "versions": self.versions,
            "depends_on_setuptools": False,
            "pypi": self.pypi,
        }

    def generate_file(self):
        env = Environment(loader=PackageLoader("pip2spack", "templates"))
        template = env.get_template("package.j2")
        return template.render(**self._generate_data())

    @staticmethod
    def generate_custom_file(abspath: str, versions):
        env = Environment(loader=FileSystemLoader(abspath))
        template = env.get_template("package.py")
        return template.render({"versions": versions})

    def download_versions(self, versions: List = None):
        if not versions:
            latest_version = self.content["info"]["version"]
            latest_filename = self.content["releases"].get(latest_version)[0][
                "filename"
            ]
            Messages.warn(
                f"Missing versions parameter, downloading a latest one v{latest_version}"
            )
            r = requests.get(url=self.url)
            self.__save_content(latest_filename, r.content)

    @staticmethod
    def __save_content(name: str, content: bytes):
        import os
        import tempfile

        path = tempfile.gettempdir()
        pip2spack = os.path.join(path, "pip2spack")
        content_filename = os.path.join(pip2spack, name)
        os.makedirs(pip2spack, exist_ok=True)

        with open(content_filename, "wb") as content_to_save:
            content_to_save.write(content)
            Messages.info(f"Content was saved into {content_filename}")



import pytest
from pytest_mock import MockFixture

from tests.pip2spack.core.mocks import (
    PYPI_CONFIGSHELL_FB_RESPONSE_MOCK,
    PYPI_CTYPESGEN_RESPONSE_MOCK,
    PYPI_RESPONSES,
)


@pytest.fixture
def mock_pypi_configshell_fb_response(mocker: MockFixture) -> None:
    mocker.patch(
        "pip2spack.core.pypi_package.PyPiPackage._download_missing_information",
        return_value=PYPI_CONFIGSHELL_FB_RESPONSE_MOCK,
    )


@pytest.fixture
def mock_pypi_jsl_response(mocker: MockFixture) -> None:
    mocker.patch(
        "pip2spack.core.pypi_package.PyPiPackage._download_missing_information",
        return_value=PYPI_RESPONSES["jsl"],
    )


@pytest.fixture
def mock_pypi_ctypesgen_response(mocker: MockFixture) -> None:
    mocker.patch(
        "pip2spack.core.pypi_package.PyPiPackage._download_missing_information",
        return_value=PYPI_CTYPESGEN_RESPONSE_MOCK["ctypesgen"],
    )



from typing import List

from pip2spack.core.pypi_package import PyPiPackage
from pip2spack.core.verification import Verification
from pip2spack.framework.helpers import create_directory, create_package
from pip2spack.framework.messages import Messages


def create_action(names: List[str]) -> None:
    ready_packages = Verification(names).available_packages

    for key, value in ready_packages.items():
        Messages.info(f"Generating package for the {key}")

        s_package = PyPiPackage(key)
        generated_file = s_package.generate_file()

        create_directory(key)
        created_package_uri = create_package(key, generated_file)
        print(
            f" OK  :: Package for {key} was generated\n\t "
            f":: If 'spack install py-{key}' return any problems, try to uncomment:\n\t "
            f":: \tdepends_on('py-setuptools', type='build')\n\t "
            f":: at {created_package_uri}"
        )



from typing import List

from pip2spack.core.pypi_package import PyPiPackage
from pip2spack.core.verification import Verification
from pip2spack.framework.messages import Messages


# TODO @NexSabre add parameter for specific version download
def download_action(names: List[str]):
    ready_packages = Verification(names).available_packages
    for key, value in ready_packages.items():
        Messages.info(f"Download latest package for the {key}")

        d_package = PyPiPackage(key)
        d_package.download_versions()



from collections import namedtuple
from http import HTTPStatus
from json import dumps

import pytest
from pytest_mock import MockFixture

from pip2spack.core.verification import Verification
from tests.pip2spack.core.mocks import PYPI_RESPONSES

Response = namedtuple("Response", "status_code content")


@pytest.fixture
def mock_requests_get_response_from_pypi_for_jsl(mocker: MockFixture) -> None:
    mocker.patch(
        "pip2spack.core.verification.requests.get",
        return_value=Response(
            status_code=HTTPStatus.OK, content=dumps(PYPI_RESPONSES["jsl"])
        ),
    )


@pytest.fixture
def mock_requests_get_404_response_from_pypi(mocker: MockFixture) -> None:
    mocker.patch(
        "pip2spack.core.verification.requests.get",
        return_value=Response(
            status_code=HTTPStatus.NOT_FOUND, content={"message": "Not Found"}
        ),
    )


class TestVerification:
    def test_verification__creation_of_object_success_story(
        self, mock_requests_get_response_from_pypi_for_jsl: None
    ) -> None:
        assert (
            Verification(
                [
                    "jsl",
                ]
            ).available_packages
            == PYPI_RESPONSES
        )

    def test_verification__negative__provide_not_existing_package_name(
        self, mock_requests_get_404_response_from_pypi
    ) -> None:
        assert not Verification(
            [
                "qwertyuiop_not_existing_package",
            ]
        ).available_packages, "Package should not exists"



import os


def get_spack_repository() -> str:
    """Return an absolute path about spack repository"""
    spack_root = os.getenv("SPACK_ROOT", "")
    if not spack_root:
        print("Please provide a information about SPACK_ROOT")
        exit(1)

    repository_suburi = os.path.relpath("var/spack/repos/builtin/packages/")
    repository_full_path = os.path.join(spack_root, repository_suburi)

    if not os.path.exists(repository_full_path):
        print("Spack repository does not exists")
        exit(1)
    return repository_full_path


def create_directory(package_name) -> bool:
    """Create a package directory for the provided package name"""
    if not package_name.startswith("py-"):
        package_name = "py-" + package_name
    target_path = os.path.join(get_spack_repository(), package_name)
    if os.path.exists(target_path):
        return False
    try:
        os.makedirs(target_path)
    except FileExistsError:
        pass
    finally:
        return os.path.exists(target_path)


def create_package(name, raw) -> str:
    if not name.startswith("py-"):
        name = "py-" + name
    target_path = os.path.join(get_spack_repository(), name)
    with open(os.path.join(target_path, "package.py"), "w") as f:
        f.write(raw)

    return os.path.join(target_path, "package.py")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pip2spack",
    version="2.0.0a1",
    author="Nex Sabre",
    author_email="nexsabre@protonmail.com",
    description="Automatically create and update a spack package base on the pypi.org information",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NexSabre/pip2spack",
    packages=setuptools.find_packages(),
    package_data={"pip2spack": ["templates/*.j2"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "requests==2.22.0",
        "jinja2==2.11.3",
        "dataclasses==0.6",
        "markupsafe==2.0.1",
        "typer==0.9.0",
    ],
    entry_points={
        "console_scripts": ["pip2spack = pip2spack.__main__:app"],
    },
)



from typing import List

from pip2spack.core.package import Package
from pip2spack.core.verification import Verification


def update_action(names: List[str]) -> None:
    package_names = Verification(names).available_packages

    if isinstance(package_names, dict):
        package_names = [k for k, v in package_names.items()]

    for p_name in package_names:
        package = Package(package_name=p_name)
        package.replace_structure_with_marker()
        package.generate_modded_package()
